"""MambuTask entity: a MambuEntity struct for Tasks.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary
"""

import copy

from .entities import MambuEntity, MambuEntityWritable


class MambuTask(MambuEntity, MambuEntityWritable):
    """MambuTask entity"""

    _prefix = "tasks"
    """prefix constant for connections to Mambu"""

    _filter_keys = [
        "username",
        "clientId",
        "groupId",
        "status",
    ]
    """allowed filters for get_all filtering"""

    _sortBy_fields = []
    """allowed fields for get_all sorting"""

    _taskLinkTypes = [
        "CLIENT",
        "GROUP",
        "BRANCH",
        "USER",
        "LOAN_ACCOUNT",
        "DEPOSIT_ACCOUNT",
        "ID_DOCUMENT",
        "LINE_OF_CREDIT",
        "GL_JOURNAL_ENTRY",
    ]

    _entities = [
        ("assignedUserKey", "mambuuser.MambuUser", "assignedUser"),
        ("taskLinkKey", "", "taskLink"),
    ]
    """3-tuples of elements and Mambu Entities"""

    def __init__(self, **kwargs):
        self._entities = copy.deepcopy(MambuTask._entities)
        super().__init__(**kwargs)

    def _assignEntObjs(
        self, entities=None, detailsLevel="BASIC", get_entities=False, debug=False
    ):
        """Overwrites `MambuPy.api.mambustruct._assignEntObjs` for MambuTask

        Determines the type of task link and instantiates accordingly
        """
        if entities is None:
            entities = self._entities

        try:
            tasklink_index = entities.index(("taskLinkKey", "", "taskLink"))
        except ValueError:
            tasklink_index = None

        if tasklink_index is not None and self.has_key("taskLinkKey"):
            if self.taskLinkType == "CLIENT":
                entities[tasklink_index] = (
                    "taskLinkKey",
                    "mambuclient.MambuClient",
                    "taskLink",
                )
            elif self.taskLinkType == "GROUP":
                entities[tasklink_index] = (
                    "taskLinkKey",
                    "mambugroup.MambuGroup",
                    "taskLink",
                )

        return super()._assignEntObjs(
            entities, detailsLevel=detailsLevel, get_entities=get_entities, debug=debug
        )

    @classmethod
    def get(cls, taskId, **kwargs):
        """get, a single task, identified by its taskId.

        Args:
          taskId (str): ID for the task

        Returns:
          instance of a task with data from Mambu
        """
        return super().get(entid=taskId, **kwargs)



"""Base and utility Classes for Mambu Objects.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary
"""

import copy


class GenericClass:  # pragma: no coverage
    """Generic class for init of MambuMapObj"""

    def __init__(self, *args, **kwargs):
        pass


class MambuMapObj:
    """An object with dictionary-like behaviour for key-value data"""

    def __init__(self, cf_class=GenericClass, **kwargs):
        self._attrs = {}
        self._cf_class = cf_class
        if hasattr(self, "_default_tzattrs"):
            self._tzattrs = copy.deepcopy(self._default_tzattrs)
        else:
            self._tzattrs = {}
        if "tzattrs" in kwargs:
            self._tzattrs.update(copy.deepcopy(kwargs.pop("tzattrs")))
        if kwargs:
            for key, val in kwargs.items():
                self._attrs[key] = val

    def __getattribute__(self, name):
        """Object-like get attribute

        When accessing an attribute, tries to find it in the _attrs
        dictionary, so now MambuMapObj may act not only as a dict-like
        structure, but as a full object-like too (this is the getter
        side).
        """
        try:
            # first, try to read 'name' as if it's a property of the object
            # if it doesn't exists as property, AttributeError raises
            return object.__getattribute__(self, name)
        except AttributeError:
            # try to read the _attrs property
            _attrs = object.__getattribute__(self, "_attrs")
            if isinstance(_attrs, list) or name not in _attrs:
                # magic won't happen when not a dict-like MambuMapObj or
                # when _attrs has not the 'name' key (this last one means
                # that if 'name' is not a property of the object too,
                # AttributeError will raise by default)
                return object.__getattribute__(self, name)
            # all else, read the property from the _attrs dict, but with a . syntax
            # if a cf_class, just return its value
            if _attrs[name].__class__.__name__ == self._cf_class.__name__:
                return _attrs[name]["value"]
            return _attrs[name]

    def __setattr__(self, name, value):
        """Object-like set attribute

        When setting an attribute, tries to set it in the _attrs
        dictionary, so now MambuMapObj acts not only as a dict-like
        structure, but as a full object-like too (this is the setter
        side).
        """
        # if name beginning with _, assign it as a property of the object
        if name[0] == "_":
            object.__setattr__(self, name, value)
        else:
            try:
                # _attrs needs to exist to make the magic happen!
                # ... if not, AttributeError raises
                _attrs = object.__getattribute__(self, "_attrs")
                if isinstance(_attrs, list):
                    # when not treating with a dict-like MambuMapObj...
                    raise AttributeError
                try:
                    # see if 'name' is currently a property of the object
                    object.__getattribute__(self, name)
                except AttributeError:
                    # if not, then assign it as a new key in the dict
                    if (
                        name in _attrs
                        and _attrs[name].__class__.__name__ == self._cf_class.__name__
                    ):
                        path = _attrs[name]["path"]
                        typecf = _attrs[name]["type"]
                        _attrs[name] = self._cf_class(value, path, typecf)
                    else:
                        _attrs[name] = value
                else:  # pragma: no cover
                    raise AttributeError
            except AttributeError:
                # all else assign it as a property of the object
                object.__setattr__(self, name, value)

    def __getitem__(self, key):
        """Dict-like key query"""
        # if a cf_class, just return its value
        if self._attrs[key].__class__.__name__ == self._cf_class.__name__:
            return self._attrs[key]["value"]
        return self._attrs[key]

    def __setitem__(self, key, value):
        """Dict-like set"""
        # if no _attrs attribute, should be automatically created?
        if (
            key in self._attrs
            and self._attrs[key].__class__.__name__ == self._cf_class.__name__
        ):
            path = self._attrs[key]["path"]
            typecf = self._attrs[key]["type"]
            self._attrs[key] = self._cf_class(value, path, typecf)
        else:
            self._attrs[key] = value

    def __delitem__(self, key):
        """Dict-like del key"""
        del self._attrs[key]

    def __str__(self):
        """Mambu object str gives a string representation of the _attrs attribute."""
        try:
            return self.__class__.__name__ + " - " + str(self._attrs)
        except AttributeError:
            return repr(self)

    def __repr__(self):
        """Mambu object repr tells the class name and the usual 'id' for it.

        If an iterable, it instead gives its length.
        """
        try:
            return self.__class__.__name__ + " - id: %s" % self._attrs["id"]
        except KeyError:
            return self.__class__.__name__ + " - " + str(self._attrs)
        except AttributeError:
            return (
                self.__class__.__name__
                + " - id: '%s' (not synced with Mambu)" % self.entid
            )
        except TypeError:
            return self.__class__.__name__ + " - len: %s" % len(self._attrs)

    def __eq__(self, other):
        """Very basic way to compare to Mambu objects.

                Only looking at their EncodedKey field (its primary key on the
                Mambu DB).

        .. todo:: a lot of improvements may be done here.
        """
        if isinstance(other, MambuMapObj):
            try:
                if "encodedKey" not in other._attrs or "encodedKey" not in self._attrs:
                    return NotImplemented
            except AttributeError:
                return NotImplemented
            return other["encodedKey"] == self["encodedKey"]

    def __hash__(self):
        """Hash of the object"""
        try:
            return hash(self.encodedKey)
        except AttributeError:
            try:
                return hash(self.__class__.__name__ + self.id)
            except Exception:
                return hash(self.__repr__())

    def __len__(self):
        """Length of the _attrs attribute.

        If dict-like (not iterable), it's the number of keys holded on the _attrs dictionary.
        If list-like (iterable), it's the number of elements of the _attrs list.
        """
        return len(self._attrs)

    def __contains__(self, item):
        """Dict-like and List-like behaviour"""
        return item in self._attrs

    def get(self, key, default=None):
        """Dict-like behaviour"""
        if isinstance(self._attrs, dict):
            return self._attrs.get(key, default)
        raise NotImplementedError  # if _attrs is not a dict

    def keys(self):
        """Dict-like behaviour"""
        try:
            return self._attrs.keys()
        except AttributeError:
            raise NotImplementedError

    def items(self):
        """Dict-like behaviour"""
        try:
            return self._attrs.items()
        except AttributeError:
            raise NotImplementedError

    def values(self):
        """Dict-like behaviour"""
        try:
            return self._attrs.values()
        except AttributeError:
            raise NotImplementedError

    def has_key(self, key):
        """Dict-like behaviour"""
        try:
            if isinstance(self._attrs, dict):
                return key in self._attrs
            raise AttributeError  # if _attrs is not a dict
        except AttributeError:  # if _attrs doesnt exist
            raise NotImplementedError



"""Schema tables for Mambu Custom fields.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary
"""

from sqlalchemy import (Column, DateTime, ForeignKey, Integer, Numeric, String,
                        Table)
from sqlalchemy.orm import backref, relationship
from sqlalchemy.orm.exc import NoResultFound

from . import schema_orm as orm

dbname = orm.dbname
session = orm.session
Base = orm.Base


class CustomFieldSet(Base):
    """CustomFieldValue table."""

    __tablename__ = "customfieldset"
    __table_args__ = {"schema": dbname, "keep_existing": True}

    # Columns
    encodedKey = Column(String, primary_key=True)
    id = Column(String)
    name = Column(String)
    type = Column(String)
    indexInList = Column(Integer)

    # Relationships
    customField = relationship("CustomField", back_populates="customFieldSet")

    def __repr__(self):
        return "<CustomFieldSet(name={}, id={})>".format(self.name, self.id)


class CustomField(Base):
    """CustomField table.
    Related with CustomFieldValue
    """

    __tablename__ = "customfield"
    __table_args__ = {"schema": dbname, "keep_existing": True}

    # Columns
    encodedKey = Column(String, primary_key=True)
    name = Column(String)
    description = Column(String)
    id = Column(String)
    type = Column(String)
    indexInList = Column(Integer)
    state = Column(String)
    valueLength = Column(String)
    dataType = Column(String)
    creationDate = Column(DateTime)
    lastModifiedDate = Column(DateTime)
    customFieldValues = relationship("CustomFieldValue", back_populates="customField")
    customFieldSet_encodedKey_oid = Column(
        String, ForeignKey(CustomFieldSet.encodedKey)
    )
    customFieldSet = relationship(CustomFieldSet, back_populates="customField")
    customFieldSelection = relationship(
        "CustomFieldSelection", back_populates="customField"
    )

    def __repr__(self):
        return "<CustomField(name={})>".format(self.name)


class CustomFieldValue(Base):
    """CustomFieldValue table."""

    __tablename__ = "customfieldvalue"
    __table_args__ = {"schema": dbname, "keep_existing": True}

    # Columns
    encodedKey = Column(String, primary_key=True)
    parentKey = Column(String)
    indexInList = Column(Integer)
    value = Column(String)
    linkedEntityKeyValue = Column(String)
    amount = Column(Numeric(50, 10))
    customFieldSetGroupIndex = Column(Integer)

    # Relationships
    customFieldKey = Column(String, ForeignKey(CustomField.encodedKey))
    customField = relationship(CustomField, back_populates="customFieldValues")
    loan = relationship(
        "LoanAccount",
        back_populates="customInformation",
        foreign_keys="CustomFieldValue.parentKey",
        primaryjoin="CustomFieldValue.parentKey == LoanAccount.encodedKey",
    )
    client = relationship(
        "Client",
        back_populates="customInformation",
        foreign_keys="CustomFieldValue.parentKey",
        primaryjoin="CustomFieldValue.parentKey == Client.encodedKey",
    )
    group = relationship(
        "Group",
        back_populates="customInformation",
        foreign_keys="CustomFieldValue.parentKey",
        primaryjoin="CustomFieldValue.parentKey == Group.encodedKey",
    )
    user = relationship(
        "User",
        back_populates="customInformation",
        foreign_keys="CustomFieldValue.parentKey",
        primaryjoin="CustomFieldValue.parentKey == User.encodedKey",
    )
    branch = relationship(
        "Branch",
        back_populates="customInformation",
        foreign_keys="CustomFieldValue.parentKey",
        primaryjoin="CustomFieldValue.parentKey == Branch.encodedKey",
    )

    @property
    def linkedclient(self):
        from .schema_clients import Client

        try:
            if self.customField.dataType == "CLIENT_LINK":
                return (
                    session.query(Client)
                    .filter(Client.encodedKey == self.linkedEntityKeyValue)
                    .one()
                )
            else:
                return None
        except NoResultFound:
            return None

    @property
    def linkedgroup(self):
        from .schema_groups import Group

        try:
            if self.customField.dataType == "GROUP_LINK":
                return (
                    session.query(Group)
                    .filter(Group.encodedKey == self.linkedEntityKeyValue)
                    .one()
                )
            else:
                return None
        except NoResultFound:
            return None

    @property
    def linkeduser(self):
        from .schema_users import User

        try:
            if self.customField.dataType == "USER_LINK":
                return (
                    session.query(User)
                    .filter(User.encodedKey == self.linkedEntityKeyValue)
                    .one()
                )
            else:
                return None
        except NoResultFound:
            return None

    def __repr__(self):
        if self.value:
            value = self.value
        elif self.linkedclient:
            value = self.linkedclient
        elif self.linkedgroup:
            value = self.linkedgroup
        elif self.linkeduser:
            value = self.linkeduser
        else:
            value = "None"

        return "<CustomFieldValue(customField={},value={})>".format(
            self.customField,
            value
        )


class CustomFieldSelection(Base):
    """CustomFieldValue table."""

    __tablename__ = "customfieldselection"
    __table_args__ = {"schema": dbname, "keep_existing": True}

    # Columns
    encodedKey = Column(String, primary_key=True)
    id = Column(String)
    value = Column(String)
    score = Column(Integer)
    selectionindex = Column(Integer)

    # Relationships
    customfieldkey = Column(String, ForeignKey(CustomField.encodedKey))
    customField = relationship(CustomField, back_populates="customFieldSelection")

    def __repr__(self):
        try:
            return "<CustomFieldSet(id={}, value={})>".format(self.id, self.value)
        except UnicodeEncodeError:
            return "<CustomFieldSet(id={}, value={})>".format(
                self.id, self.value.encode("utf8")
            )



# coding: utf-8
"""Mambu Activity objects - DEPRECATED LEGACY MODULE.

.. warning::

   DEPRECATED: This module uses Mambu API v1 which is no longer supported.
   This code is maintained only for backwards compatibility with legacy systems.
   It will be removed in a future version of MambuPy.

MambuActivity holds a mambu activity. Don't instantiate this class
directly. Look at MambuActivity pydoc for further info.

MambuActivities holds a list of activities.
"""
import json
import warnings
from copy import deepcopy
from datetime import datetime

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from ..mambugeturl import getmambuurl
from ..mambuutil import (
    OUT_OF_BOUNDS_PAGINATION_LIMIT_VALUE,
    MambuCommError,
    MambuError,
    apipwd,
    apiuser,
    encoded_dict,
    iri_to_uri,
    setup_logging,
    strip_tags,
)

warnings.warn(
    "MambuPy.rest.mambuactivity is deprecated and will be removed in a future version. "
    "Mambu API v1 is no longer supported by Mambu.",
    DeprecationWarning,
    stacklevel=2,
)

logger = setup_logging(__name__)


# =============================================================================
# Utility Classes (from mamburestutils.py)
# =============================================================================


class RequestsCounter(object):
    """Singleton that counts requests.

    If you want to control the number of requests you do to Mambu, you
    may find this counter useful. Since it is a Singleton, every Mambu
    object shares it and increases the amount of requests counted here,
    so you may read it on every Mambu object you have per Python
    session you hold.
    """

    __instance = None

    def __new__(cls, *args, **kwargs):
        if not cls.__instance:
            cls.__instance = super(RequestsCounter, cls).__new__(cls, *args, **kwargs)
            cls.requests = []
            cls.cnt = 0
        return cls.__instance

    def add(cls, temp):
        cls.requests.append(temp)
        cls.cnt += 1

    def reset(cls):
        cls.cnt = 0
        cls.requests = []


class MambuStructIterator:
    """Enables iteration for Mambu objects that hold lists."""

    def __init__(self, wrapped):
        self.wrapped = wrapped
        self.offset = 0

    def __next__(self):
        if self.offset >= len(self.wrapped):
            raise StopIteration
        else:
            item = self.wrapped[self.offset]
            self.offset += 1
            return item

    def next(self):
        return self.__next__()

    def __iter__(self):
        return self


# =============================================================================
# URL Function (from mambugeturl.py)
# =============================================================================


def _get_parameters_url(kwargs):
    """Build URL parameters from kwargs."""
    getparams = []
    for k, v in kwargs.items():
        if v is not None:
            getparams.append("%s=%s" % (k, v))
    return getparams


def getactivitiesurl(dummyId="", *args, **kwargs):
    """Request Activities URL.

    dummyId is used because MambuStruct always requires an Id from an
    entity, but the Mambu API doesn't requires it for Activities,
    because of that dummyId defaults to '', but in practice it is never
    used.

    Currently implemented filter parameters:
    * from
    * to
    * branchID
    * clientId
    * centreID
    * userID
    * loanAccountId
    * groupId
    * limit

    See Mambu official developer documentation for further details.

    .. warning:: API V2: NOT compatible. Apparently not yet implemented.
    """
    getparams = []
    if kwargs:
        try:
            getparams.append("from=%s" % kwargs["fromDate"])
            del kwargs["fromDate"]
        except Exception:
            getparams.append("from=%s" % "1900-01-01")

        try:
            getparams.append("to=%s" % kwargs["toDate"])
            del kwargs["toDate"]
        except Exception:
            hoy = datetime.now().strftime("%Y-%m-%d")
            getparams.append("to=%s" % hoy)

        getparams.extend(_get_parameters_url(kwargs))

    url = (
        getmambuurl(*args, **kwargs)
        + "activities"
        + ("" if len(getparams) == 0 else "?" + "&".join(getparams))
    )
    return url


mod_urlfunc = getactivitiesurl


# =============================================================================
# Minimal MambuStruct Base Class
# =============================================================================


class MambuStruct(object):
    """Minimal MambuStruct base for Activities.

    This is a stripped-down version containing only the functionality
    required for MambuActivity and MambuActivities.
    """

    RETRIES = 5

    @staticmethod
    def serialize_fields(data):
        """Turns every attribute of the Mambu object into a string representation."""
        if isinstance(data, MambuStruct):
            return data.serialize_struct()
        try:
            it = iter(data)
        except TypeError:
            return str(data)
        if type(it) == type(iter([])):
            return [MambuStruct.serialize_fields(e) for e in it]
        elif type(it) == type(iter({})):
            return {k: MambuStruct.serialize_fields(data[k]) for k in it}
        return str(data)

    def __getitem__(self, key):
        """Dict-like key query"""
        return self.attrs[key]

    def __setitem__(self, key, value):
        """Dict-like set"""
        self.attrs[key] = value

    def __delitem__(self, key):
        """Dict-like del key"""
        del self.attrs[key]

    def __hash__(self):
        """Hash of the object"""
        try:
            return hash(self.id)
        except Exception:
            return hash(self.__repr__())

    def __getattribute__(self, name):
        """Object-like get attribute"""
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            attrs = object.__getattribute__(self, "attrs")
            if type(attrs) == list or name not in attrs:
                return object.__getattribute__(self, name)
            return attrs[name]

    def __setattr__(self, name, value):
        """Object-like set attribute"""
        try:
            attrs = object.__getattribute__(self, "attrs")
            if type(attrs) == list:
                raise AttributeError
            try:
                object.__getattribute__(self, name)
            except AttributeError:
                attrs[name] = value
            else:
                raise AttributeError
        except AttributeError:
            object.__setattr__(self, name, value)

    def __repr__(self):
        """Mambu object repr tells the class name and the usual 'id' for it."""
        try:
            return self.__class__.__name__ + " - id: %s" % self.attrs["id"]
        except KeyError:
            return self.__class__.__name__ + " (no standard entity)"
        except AttributeError:
            return (
                self.__class__.__name__
                + " - id: '%s' (not synced with Mambu)" % self.entid
            )
        except TypeError:
            return self.__class__.__name__ + " - len: %s" % len(self.attrs)

    def __str__(self):
        """Mambu object str gives a string representation of the attrs attribute."""
        try:
            return self.__class__.__name__ + " - " + str(self.attrs)
        except AttributeError:
            return repr(self)

    def __len__(self):
        """Length of the attrs attribute."""
        return len(self.attrs)

    def __eq__(self, other):
        """Compare two Mambu objects by EncodedKey."""
        if isinstance(other, MambuStruct):
            try:
                if "encodedKey" not in other.attrs or "encodedKey" not in self.attrs:
                    return NotImplemented
            except AttributeError:
                return NotImplemented
            return other["encodedKey"] == self["encodedKey"]

    def has_key(self, key):
        """Dict-like behaviour"""
        try:
            if type(self.attrs) == dict:
                return key in self.attrs
            else:
                raise AttributeError
        except AttributeError:
            raise NotImplementedError

    def __contains__(self, item):
        """Dict-like and List-like behaviour"""
        return item in self.attrs

    def get(self, key, default=None):
        """Dict-like behaviour"""
        if type(self.attrs) == dict:
            return self.attrs.get(key, default)
        else:
            raise NotImplementedError

    def keys(self):
        """Dict-like behaviour"""
        try:
            return self.attrs.keys()
        except AttributeError:
            raise NotImplementedError

    def items(self):
        """Dict-like behaviour"""
        try:
            return self.attrs.items()
        except AttributeError:
            raise NotImplementedError

    def init(self, attrs={}, *args, **kwargs):
        """Initialize from a dictionary responded by Mambu."""
        self.attrs = attrs
        self.preprocess()
        self.convert_dict_to_attrs(*args, **kwargs)
        self.postprocess()
        try:
            self.entid = self.id
        except (TypeError, AttributeError):
            pass
        try:
            for meth in kwargs["methods"]:
                try:
                    getattr(self, meth)()
                except Exception:  # pragma: no cover
                    # Defensive: silently ignore method execution errors
                    pass
        except Exception:
            pass
        try:
            for propname, propval in kwargs["properties"].items():
                setattr(self, propname, propval)
        except Exception:
            pass

    def serialize_struct(self):
        """Makes a string from each element on the attrs attribute."""
        return MambuStruct.serialize_fields(self.attrs)

    def __copy_args(self, args, kwargs):
        self.__args = deepcopy(args)
        for k, v in kwargs.items():
            self.__kwargs[k] = deepcopy(v)

    def __init__(self, urlfunc, entid="", *args, **kwargs):
        """Initialize a new Mambu object."""
        self.entid = entid
        self.rc = RequestsCounter()

        connect = False

        try:
            self.__debug = kwargs.pop("debug")
        except KeyError:
            self.__debug = False

        try:
            self.__formato_fecha = kwargs.pop("date_format")
        except KeyError:
            self.__formato_fecha = "%Y-%m-%dT%H:%M:%S+0000"

        try:
            self.__data = kwargs.pop("data")
        except KeyError:
            self.__data = None

        try:
            self.__method = kwargs.pop("method")
        except KeyError:
            if self.__data:
                self.__method = "POST"
            else:
                self.__method = "GET"

        try:
            self.__limit = kwargs["limit"]
            self.__inilimit = self.__limit
            kwargs.pop("limit")
        except KeyError:
            self.__limit = 0

        try:
            self.__offset = kwargs["offset"]
            kwargs.pop("offset")
        except KeyError:
            self.__offset = 0

        try:
            self.custom_field_name = kwargs.pop("custom_field_name")
        except KeyError:
            pass

        try:
            self.__user = kwargs.pop("user")
        except (KeyError, AttributeError):
            self.__user = apiuser
        try:
            self.__pwd = kwargs.pop("pwd")
        except (KeyError, AttributeError):
            self.__pwd = apipwd

        self.__headers = {"Accept": "application/vnd.mambu.v1+json"}

        try:
            if kwargs.pop("connect"):
                connect = True
        except KeyError:
            connect = True

        self.__args = ()
        self.__kwargs = {}

        self.__copy_args(args, kwargs)

        self.__urlfunc = urlfunc
        if self.__urlfunc is None:
            return

        if connect:
            self.connect(*args, **kwargs)

    def __request(self, url, user, pwd):
        """Make the HTTP request to Mambu."""
        retry_strategy = Retry(
            total=MambuStruct.RETRIES,
            status_forcelist=[429, 502, 503, 504],
            backoff_factor=1,
            allowed_methods=["PUT", "DELETE", "POST", "PATCH"],
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        http = requests.Session()
        http.mount("https://", adapter)
        http.mount("http://", adapter)

        logger.debug(
            "%s %s, about to make %s request: url %s, data %s, headers %s",
            self.__class__.__name__,
            self.entid,
            self.__method,
            url,
            self.__data,
            [(k, v) for k, v in self.__headers.items()],
        )

        if self.__data:
            data = json.dumps(encoded_dict(self.__data))
            self.__headers["Content-Type"] = "application/json"
            if self.__method == "PATCH":
                resp = http.patch(
                    url, data=data, headers=self.__headers, auth=(user, pwd)
                )
            else:
                resp = http.post(
                    url, data=data, headers=self.__headers, auth=(user, pwd)
                )
        else:
            if self.__method == "DELETE":
                resp = http.delete(url, headers=self.__headers, auth=(user, pwd))
            else:
                resp = http.get(url, headers=self.__headers, auth=(user, pwd))

        self.rc.add(datetime.now())

        logger.debug(
            "%s %s,response code to %s: %s",
            self.__class__.__name__,
            self.entid,
            self.__method,
            resp,
        )

        return resp

    def __process_jsonresp(self, jsonresp, limit, jsresp):
        window = True
        if type(jsonresp) == list:
            try:
                jsresp.extend(jsonresp)
            except AttributeError:
                jsresp = jsonresp
            if len(jsonresp) < limit:
                window = False
        else:
            jsresp = jsonresp
            self._is_mambu_error(jsresp)
            window = False
        return jsresp, window

    def __request_and_process(self, jsresp, url, user, pwd, limit, offset):
        resp = ""
        try:
            resp = self.__request(url, user, pwd)
            resp.raise_for_status()

            self._raw_response_data = resp.content
            jsonresp = json.loads(resp.content)
            jsresp, window = self.__process_jsonresp(jsonresp, limit, jsresp)
        except requests.exceptions.HTTPError as httpex:
            logger.error(
                "%s %s, HTTPError on %s request: url %s",
                self.__class__.__name__,
                self.entid,
                self.__method,
                url,
            )
            try:
                jsresp = json.loads(resp.content)
            except ValueError:
                cont = resp.content
                raise MambuError(
                    strip_tags(cont.decode())
                    .strip()
                    .replace("\n\n", ": ", 1)
                    .replace("\n", ". ", 1)
                    .replace("\n", " ")
                )
            except Exception as ex:  # pragma: no cover
                # Defensive: json.loads raises ValueError for invalid JSON
                raise MambuError("JSON Error: %s" % repr(ex))

            try:
                self._is_mambu_error(jsresp)
            except AttributeError:  # pragma: no cover
                # Defensive: jsresp always dict after json.loads
                pass
        except requests.exceptions.RetryError as rerr:
            raise MambuCommError(
                "ERROR I can't communicate with Mambu: {}".format(str(rerr))
            )
        except Exception as ex:
            logger.exception(
                "%s %s, Exception: %s", self.__class__.__name__, self.entid, str(ex)
            )
            raise ex

        return jsresp, window

    def _is_mambu_error(self, jsresp):
        if (
            "returnCode" in jsresp
            and "returnStatus" in jsresp
            and jsresp["returnCode"] != 0
        ):
            raise MambuError(jsresp["returnStatus"])
        return False

    def __credentials(self):
        try:
            user = self.__user
        except AttributeError:  # pragma: no cover
            # Defensive: __user always set in __init__
            self.__user = apiuser
            user = self.__user
        try:
            pwd = self.__pwd
        except AttributeError:  # pragma: no cover
            # Defensive: __pwd always set in __init__
            self.__pwd = apipwd
            pwd = self.__pwd
        return user, pwd

    def connect(self, *args, **kwargs):
        """Connect to Mambu and make the REST API request."""
        self.__copy_args(args, kwargs)

        if not self.__urlfunc:
            return

        offset = self.__offset
        window = True
        jsresp = {}
        while window:
            if not self.__limit or self.__limit > OUT_OF_BOUNDS_PAGINATION_LIMIT_VALUE:
                limit = OUT_OF_BOUNDS_PAGINATION_LIMIT_VALUE
            else:
                limit = self.__limit

            user, pwd = self.__credentials()

            urlfunc_params = self.__kwargs.copy()
            if self.__method == "GET" and not self.__data:
                urlfunc_params["limit"] = limit
                urlfunc_params["offset"] = offset
            url = iri_to_uri(
                self.__urlfunc(str(self.entid), *self.__args, **urlfunc_params)
            )
            self.__url = url

            jsresp, window = self.__request_and_process(
                jsresp, url, user, pwd, limit, offset
            )

            offset = offset + limit
            if self.__limit:
                self.__limit -= limit
                if self.__limit <= 0:
                    window = False
                    self.__limit = self.__inilimit

        if ("documents" not in url) and self.__method not in ["PATCH", "DELETE"]:
            if type(iter(jsresp)) == type(iter([])):
                logger.debug(
                    "%s, url %s, %s retrieved",
                    self.__class__.__name__,
                    url,
                    len(jsresp),
                )
            self.init(attrs=jsresp, *self.__args, **self.__kwargs)

    def _process_fields(self):
        """Default info massage to appropriate format/style."""
        try:
            for k, v in self.items():
                try:
                    self[k] = v.strip()
                except Exception:
                    pass
        except NotImplementedError:
            pass

    def preprocess(self):
        """Preprocess hook."""
        self._process_fields()

    def postprocess(self):
        """Postprocess hook."""
        self._process_fields()

    def _convert_data_to_pytype(self, data):
        """Convert data to Python built-in types."""
        try:
            d = int(data)
            if str(d) != data:
                return data
            return d
        except (TypeError, ValueError):
            try:
                return float(data)
            except (TypeError, ValueError):
                try:
                    return self.util_date_format(data)
                except (TypeError, ValueError):
                    return data
        return data

    def _convert_dict_to_pytypes(self, data):
        """Recursively convert fields to Python objects."""
        constantFields = [
            "id",
            "groupName",
            "name",
            "homePhone",
            "mobilePhone1",
            "phoneNumber",
            "postcode",
            "emailAddress",
            "description",
        ]
        try:
            it = iter(data)
            if type(it) == type(iter({})):
                d = {}
                for k in it:
                    if k in constantFields:
                        d[k] = data[k]
                    else:
                        d[k] = self._convert_dict_to_pytypes(data[k])
                data = d
            if type(it) == type(iter([])):
                data = [self._convert_dict_to_pytypes(e) for e in it]
        except TypeError:
            pass
        except Exception as ex:  # pragma: no cover
            # Defensive: re-raise unexpected exceptions
            raise ex
        return self._convert_data_to_pytype(data)

    def convert_dict_to_attrs(self, *args, **kwargs):
        """Convert each element on attrs to a proper Python object."""
        self.attrs = self._convert_dict_to_pytypes(self.attrs)

    def util_date_format(self, field, formato=None):
        """Convert a datetime field to a datetime using specified format."""
        if not formato:
            try:
                formato = self.__formato_fecha
            except AttributeError:
                formato = "%Y-%m-%dT%H:%M:%S+0000"
        return datetime.strptime(
            datetime.strptime(field, "%Y-%m-%dT%H:%M:%S+0000").strftime(formato),
            formato,
        )


# =============================================================================
# Activity Classes
# =============================================================================


class MambuActivity(MambuStruct):
    """A loan Activity from Mambu.

    Don't instantiate this class directly. It's mostly used by
    MambuActivities to configure each of its elements as
    MambuActivity objects.
    """

    def __init__(self, urlfunc=mod_urlfunc, entid="", *args, **kwargs):
        """Initialize the MambuStruct."""
        MambuStruct.__init__(self, urlfunc, entid, *args, **kwargs)

    def __repr__(self):
        """Show the activityid of the activity."""
        try:
            return self.__class__.__name__ + " - activityid: %s" % self["activity"]
        except KeyError:
            return self.__class__.__name__ + " (no activity key)"


class MambuActivities(MambuStruct):
    """A list of loan Activities from Mambu."""

    def __init__(self, urlfunc=mod_urlfunc, entid="", *args, **kwargs):
        MambuStruct.__init__(self, urlfunc, entid, *args, **kwargs)

    def __iter__(self):
        return MambuStructIterator(self.attrs)

    def convert_dict_to_attrs(self, *args, **kwargs):
        """Convert each element of the list to a MambuActivity object."""
        for n, a in enumerate(self.attrs):
            try:
                params = self.params
            except AttributeError:
                params = {}
            kwargs.update(params)
            try:
                self.mambuactivityclass
            except AttributeError:
                self.mambuactivityclass = MambuActivity

            activity = self.mambuactivityclass(urlfunc=None, entid=None, *args, **kwargs)
            activity.init(a, *args, **kwargs)
            activity._MambuStruct__urlfunc = getactivitiesurl
            self.attrs[n] = activity



"""Basic Struct for Mambu Objects.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary
"""

import copy
from datetime import datetime
from importlib import import_module

from .classes import MambuMapObj
from ..mambuutil import MambuPyError, date_format


class MambuStruct(MambuMapObj):
    """Basic Struct for Mambu Objects with basic connection functionality."""

    _attrs = {}
    """Properties of the Mambu object"""

    _tzattrs = {}
    """TimeZones info:

        `_convertDict2Attrs` loses TZ info on datetime fields.

        We will save them on `_tzattrs` field. Prefering this method since this
        allows comparison with datetimes on your code without needing TZ
        initialized.

        For example:

        >>> today = datetime.now()
        >>> loan.creationDate < today

        That code works since now() doesn't have any TZ info.

        If creationDate has TZ info in it, you could not make the comparison.

        So, we need to preserve the TZ info somewhere else.

        Since TZ info is NOT the same for ALL the Mambu instance (each datetime
        may have different TZs due to daylight saving time differences for
        example), _tzattrs will hold the TZ for each datetime field.

        Only-date fields are also included but they lack TZ info by definition,
        so those fileds get a None TZ info.
    """

    _as_utc = False
    """When True, all datetimes will be converted as UTC in _convertDict2Attrs."""

    _vos = []
    """List of Value Objects in the struct's _attrs.

       Each element must be a 2-tuple:
           ("name_in_attrs", "vos.class to instantiate")

       `_extractVOs` loops this list to instantiate the corresponding Value
       Objects inside _attrs. If the element in _attrs happens to be a list,
       the result will be a list of instantiated Value Objects

       `_updateVOs` loops this list to update the corresponding data in _attrs
    """

    _entities = []
    """List of Entities in the struct's _attrs.

    Each element must be a 3-tuple:
        ("name_of_key_in_attrs", # element in attrs to obtain the key
         "module.class to instantiate", # class to instantiate for the given key
         "name_of_obj_in_attrs") # where in attrs to update with instantiated obj
    """

    def __getattribute_for_one_cf(self, ent, value, mcf):
        if mcf and isinstance(value, str) and "_LINK" in mcf.type:
            link_type = mcf.type[:-5]
            package = "mambu" + link_type.lower()
            classname = "Mambu" + link_type.capitalize()
            classpath = package + "." + classname

            return lambda **kwargs: self.getEntities(
                [ent], config_entities=[(ent, classpath, ent)], **kwargs
            )[0]

        return lambda **kwargs: value

    def __getattribute_for_loop(self, ent, value, mcf):
        attributes = []
        for ind, item in enumerate(value):
            values = {}
            for key, val in item.items():
                values[key] = self.__getattribute_for_one_cf(
                    ent + "/" + str(ind) + "/" + key, val, mcf[key]
                )()
            attributes.append(values)
        return lambda **kwargs: attributes

    def __getattribute_for_cf(self, ent, entity):
        if isinstance(entity.value, list):
            return self.__getattribute_for_loop(ent, entity.value, entity.mcf)
        return self.__getattribute_for_one_cf(ent, entity.value, entity.mcf)

    def __getattribute__(self, name):
        """Object-like get attribute for MambuStructs.

        If the attribute is not present at the _attrs dict, tries to build a
        function that calls :py:meth:`getEntities`, which in turns will try
        to instantiate an entity according to the requested attribute name
        """
        try:
            return super().__getattribute__(name)
        except AttributeError as attr_err:
            if name[0:4] == "get_" and len(name) > 4:
                ent = name[4:]
                if ent in self._attrs.keys():
                    entity = self._attrs[ent]
                    if entity.__class__.__name__ == self._cf_class.__name__:
                        if not entity.mcf:
                            entity.get_mcf()
                        return self.__getattribute_for_cf(ent, entity)
                    return lambda **kwargs: entity
                return lambda **kwargs: self.getEntities([ent], **kwargs)[0]
            else:
                raise attr_err

    def __convert_from_dict_to_basic_types_non_constant_fields(
        self, k, data_dict, data, tzdata, constantFields
    ):
        try:
            data_dict[k] = self.__convert_to_basic_types(data[k], tzdata[k], constantFields)
            if type(data_dict[k]) not in [dict, list, datetime]:
                del tzdata[k]
            elif isinstance(data_dict[k], datetime):
                tzdata[k] = datetime.fromisoformat(tzdata[k]).tzname()
        except (KeyError, ValueError, TypeError):
            data_dict[k] = self.__convert_to_basic_types(data[k])

    def __convert_from_dict_to_basic_types(self, it_dict, data, tzdata, constantFields):
        data_dict = {}
        for k in it_dict:
            if k in constantFields or (len(k) > 2 and k[-3:] == "Key"):
                data_dict[k] = data[k]
                if tzdata and k in tzdata:
                    del tzdata[k]
            else:
                self.__convert_from_dict_to_basic_types_non_constant_fields(
                    k, data_dict, data, tzdata, constantFields
                )
        return data_dict

    def __convert_from_list_to_basic_types(self, it_list, data, tzdata, constantFields):
        data_list = []
        for num, (e, te) in enumerate(zip(it_list, tzdata)):
            d = self.__convert_to_basic_types(e, te, constantFields)
            if type(d) not in [dict, list, datetime]:
                tzdata[num] = None
            elif isinstance(d, datetime):
                tzdata[num] = datetime.fromisoformat(tzdata[num]).tzname()
            data_list.append(d)
        return data_list

    def __convert_to_basic_types_base_cases(self, data):
        try:
            if data in ["TRUE", "true", "FALSE", "false"]:
                return data.lower() == "true"
        except TypeError:
            # some data types override __eq__ and may not be compared to str
            # in such cases continue evaluating down under
            pass
        try:
            i_data = int(data)
            if (
                str(i_data) != data
            ):  # if string has trailing 0's, leave it as string, to not lose them
                return data
            return i_data
        except (TypeError, ValueError):
            try:
                f_data = float(data)
                return f_data
            except (TypeError, ValueError):
                try:
                    return date_format(data, as_utc=self._as_utc)
                except (TypeError, ValueError):
                    return data

        return data

    def __convert_to_basic_types(self, data, tzdata=None, constantFields=None):
        """Recursively convert the fields on the data given to a python
        object.

        If data is iterable, iterates its elements and try to convert them.

        If data is a string, tries to convert its value to a basic data type:

        Basic data types supported:
          - int: an int number
          - float: a floating point number
          - datetime: if the string holds a valid datetime in a date_format
                      specific format. Considers UTC if _as_utc is True.

        A list of fields that should stay as-they-come (strings) is supported.
        All fields whose name ends with "Key" is also ignored.

        Args:
          data (obj): an object whose value should be converted to a basic
                      type
          tzdata (obj): mirror of data, holding only the TZ data for datetimes.
          constantFields (list): fields that will be ignored for conversion
        """
        if not constantFields:
            constantFields = []
        # Iterators, lists and dictionaries
        # Here comes the recursive calls!
        try:
            it = data
            if isinstance(it, dict):
                data = self.__convert_from_dict_to_basic_types(
                    it, data, tzdata, constantFields
                )
            if isinstance(it, MambuMapObj):
                it._convertDict2Attrs()
            if isinstance(it, list):
                data = self.__convert_from_list_to_basic_types(
                    it, data, tzdata, constantFields
                )
        except TypeError:
            pass
        except Exception as ex:  # pragma: no cover
            # unknown exception
            raise ex

        # Base case!
        return self.__convert_to_basic_types_base_cases(data)

    def _convertDict2Attrs(self, *args, **kwargs):
        """Each element on the attrs attribute gets converted to a
        proper python object, depending on type.

        Some default constantFields are left as is (strings), because they are
        better treated as strings. This includes any field whose name ends with
        'Key'.
        """
        constantFields = [
            "id",
            "groupName",
            "name",
            "homePhone",
            "mobilePhone",
            "mobilePhone2",
            "postcode",
            "emailAddress",
            "description",
        ]
        # and any field whose name ends with "Key"

        self._attrs = self.__convert_to_basic_types(
            self._attrs, self._tzattrs, constantFields
        )

    def __convert_from_dict_basic_types_to_str(self, it_dict, data, tzdata):
        d = {}
        for k in it_dict:
            if tzdata and k in tzdata:
                d[k] = self.__convert_basic_types_to_str(data[k], tzdata[k])
            else:
                d[k] = self.__convert_basic_types_to_str(data[k])
        return d

    def __convert_from_list_basic_types_to_str(self, it_list, data, tzdata):
        data_list = []
        if tzdata:
            for e, te in zip(it_list, tzdata):
                data_list.append(self.__convert_basic_types_to_str(e, te))
        else:
            for e in it_list:
                data_list.append(self.__convert_basic_types_to_str(e))
        return data_list

    def __convert_basic_types_to_str_base_cases(self, data, tzdata):
        if isinstance(data, datetime):
            data_asdate = data.isoformat()
            if tzdata:  # no tzdata means a date (no time) object
                data_asdate += tzdata[-6:]
            else:
                data_asdate = data_asdate[:10]
            return data_asdate
        if data in [True, False]:
            return str(data).lower()
        return str(data)

    def __convert_basic_types_to_str(self, data, tzdata=None):
        """Recursively convert the fields on the data given to strings.

        Datetime objects are converted to ISO format, adding TZ info if
        available on tzdata.

        Skips every MambuMapObj owned by this entity.

        If the object is an iterable one, it goes down to each of its
        elements and turns its attributes too

        The base case is when it's a MambuMapObj class (this one) so it
        just 'serializes' the attr atribute.

        Args:
          data (obj): an object whose value should be converted to string.
          tzdata (obj): mirror of data, holding only the TZ data for datetimes.
        """
        # Base case!
        if isinstance(data, MambuMapObj):
            data._serializeFields()
            return data
        try:
            it = iter(data)
        except TypeError:
            return self.__convert_basic_types_to_str_base_cases(data, tzdata)

        # Recursive calls
        if isinstance(data, list):
            return self.__convert_from_list_basic_types_to_str(it, data, tzdata)
        elif isinstance(data, dict):
            return self.__convert_from_dict_basic_types_to_str(it, data, tzdata)
        # elif ... tuples? sets?
        return data

    def _serializeFields(self, *args, **kwargs):
        """Every attribute of the Mambu object is turned in to a string
        representation.

        Args:
          data (obj): an object whose value should be converted to string.
          tzdata (obj): mirror of data, holding only the TZ data for datetimes.
        """
        self._attrs = self.__convert_basic_types_to_str(self._attrs, self._tzattrs)

    def __extract_customfields_from_dict(self, val_dict, attr, attrs):
        for key, value in val_dict.items():
            attrs[key] = self._cf_class(value, "/{}/{}".format(attr, key), "STANDARD")

    def __extract_customfields_from_list(self, val_list, attr, attrs):
        attrs[attr[1:]] = self._cf_class(
            copy.deepcopy(val_list), "/{}".format(attr), "GROUPED"
        )
        for ind, value in enumerate(val_list):
            if isinstance(value, dict):
                for key, subvalue in value.items():
                    if key[0] != "_":
                        mecf = self._cf_class(
                            subvalue,
                            "/{}/{}/{}".format(attr, ind, key),
                            "GROUPED",
                        )
                        attrs[key + "_" + str(ind)] = mecf
                        # attrs[attr[1:]][ind][key] = mecf
            else:
                raise MambuPyError(
                    "CustomFieldSet {} is not a list of dictionaries!".format(attr)
                )

    def _extractCustomFields(self, attrs=None):
        """Loops through every custom field set and extracts custom field values
        on the root of the attrs argument.

        Args:
          attrs (dict): dictionary of fields from where customfields will be
                        extracted. If None, `self._attrs` will be used
        """
        if not attrs:
            attrs = self._attrs

        for attr, val in [atr for atr in attrs.items() if atr[0][0] == "_"]:
            if isinstance(val, dict):
                self.__extract_customfields_from_dict(val, attr, attrs)
            elif isinstance(val, list):
                self.__extract_customfields_from_list(val, attr, attrs)
            else:
                raise MambuPyError("CustomFieldSet {} is not a dictionary!".format(attr))

    def __get_value_or_ek(self, obj):
        if isinstance(obj, MambuStruct):
            return obj.encodedKey
        return obj

    def __update_customfields_from_dict(self, val_dict, attr, cfs):
        for key in val_dict.keys():
            try:
                if self[key] in [True, False]:
                    self[key] = str(self[key]).upper()
                self._attrs[attr][key] = self.__get_value_or_ek(self[key])
                cfs.append(key)
            except KeyError:
                pass

    def __update_customfields_from_list(self, val_list, attr, cfs):
        for ind, value in [
            (ind, value) for ind, value in enumerate(val_list) if isinstance(value, dict)
        ]:
            for key in [k for k in value.keys() if k[0] != "_"]:
                try:
                    if self[key + "_" + str(ind)] in [True, False]:
                        self[key + "_" + str(ind)] = str(
                            self[key + "_" + str(ind)]
                        ).upper()
                    if self[attr][ind][key] != self[key + "_" + str(ind)]:
                        self[attr[1:]][ind][key] = self.__get_value_or_ek(
                            self[key + "_" + str(ind)]
                        )
                    cfs.append(key + "_" + str(ind))
                except KeyError:
                    pass
                except TypeError:
                    cfs.append(key + "_" + str(ind))
        try:
            self._attrs[attr] = copy.deepcopy(self[attr[1:]])
            cfs.append(attr[1:])
        except KeyError:
            pass

    def _updateCustomFields(self):
        """Loops through every custom field set and update custom field values
        with the corresponding property at the root of the `_attrs` dict, then
        deletes the property at root"""
        cfs = []
        # updates customfieldsets
        for attr, val in [
            atr
            for atr in self._attrs.items()
            if atr[0][0] == "_"
            and atr[0][1:] not in [ent_config[2] for ent_config in self._entities]
        ]:
            if isinstance(val, dict):
                self.__update_customfields_from_dict(val, attr, cfs)
            elif isinstance(val, list):
                self.__update_customfields_from_list(val, attr, cfs)
            else:
                raise MambuPyError(
                    "CustomFieldSet {} is not a dictionary or list of dictionaries!".format(
                        attr
                    )
                )
        # deletes _attrs root keys of custom fields
        for field in cfs:
            del self._attrs[field]

    def __extract_vos_from_list(
        self, vo_data, vos_module, voclass, elem, get_entities=False, debug=False
    ):
        """Extracts the VOs from a list.

        Given a list of data representing a VO, extract and instantiate each of
        them.

        Args:
          vo_data (list): list of VOs to extract
          vos_module (module): VOs module from MambuPy
          voclass (class): class of the VOs to instantiate
          elem (obj): elem at attrs where the list comes from
          get_entities (bool): should MambuPy automatically instantiate other
                               MambuPy entities found inside the Value Objects?
          debug (bool): print debugging info

        Returns:
          (vo_obj, already) (tuple): (vo obj, bool) a list of instances of VOs
                                     and if it has already been instantiated
                                     (bool)
        """
        vo_obj = []
        already = False
        for ind, item in enumerate(vo_data):
            if isinstance(item, getattr(vos_module, voclass)):
                already = True
                continue
            vo_item = getattr(vos_module, voclass)(**item)
            vo_item._tzattrs = self._tzattrs[elem][ind]
            vo_item._extractVOs()
            if get_entities:
                vo_item._assignEntObjs(
                    vo_item._entities, get_entities=get_entities, debug=debug
                )
            vo_obj.append(vo_item)

        return vo_obj, already

    def _extractVOs(self, get_entities=False, debug=False):
        """Loops _vos list to instantiate the corresponding Value Objects.

        If the element in _attrs happens to be a list, the result will be a
        list of instantiated Value Objects.

        End result, the key with the name of the element will be replaced with
        the instantiated Value Object. And the original element will change its
        key name from 'elem' to 'vo_elem'.

        Args:
          get_entities (bool): should MambuPy automatically instantiate other
                               MambuPy entities found inside the Value Objects?
          debug (bool): print debugging info
        """
        vos_module = import_module(".vos", "mambupy.api")
        for elem, voclass in self._vos:
            try:
                vo_data = self._attrs[elem]
            except KeyError:
                continue

            if isinstance(vo_data, list):
                vo_obj, already = self.__extract_vos_from_list(
                    vo_data, vos_module, voclass, elem, get_entities, debug
                )
                if already:
                    continue
            elif isinstance(vo_data, getattr(vos_module, voclass)):
                continue
            else:
                vo_obj = getattr(vos_module, voclass)(**vo_data)
                vo_obj._tzattrs = self._tzattrs[elem]
                vo_obj._extractVOs()
                if get_entities:
                    vo_obj._assignEntObjs(
                        vo_obj._entities, get_entities=get_entities, debug=debug
                    )

            self._attrs[elem] = vo_obj

    def __update_vos_from_list(self, vo_obj, vos_module, voclass):
        """Updates the VOs from a list.

        Given a list of VOs, updates each element's data to the attrs dict.

        Args:
          vo_obj (obj): the VO which data will be updated at the attrs dict.
          vos_module (module): VOs module from MambuPy
          voclass (class): class of the VOs to instantiate

        Returns:
          (vo_data, already) (tuple): (dict, bool) a list of updated data and
                                      if it has already been updated (bool)
        """
        vo_data = []
        already = False
        for item in vo_obj:
            if not isinstance(item, getattr(vos_module, voclass)):
                already = True
                continue
            item._updateVOs()
            vo_item = copy.deepcopy(item._attrs)
            vo_data.append(vo_item)

        return vo_data, already

    def _updateVOs(self):
        """Loops _vos list to update the corresponding data in _attrs

        End result, the element with the original element in _attr will have
        its data updated, the Value Object will dissappear and the key name
        of the original element will return to be from 'vo_elem' to 'elem'
        """
        vos_module = import_module(".vos", "mambupy.api")
        for elem, voclass in self._vos:
            try:
                vo_obj = self._attrs[elem]
            except KeyError:
                continue
            if isinstance(vo_obj, list):
                vo_data, already = self.__update_vos_from_list(
                    vo_obj, vos_module, voclass
                )
                if already:
                    continue
            elif not isinstance(vo_obj, getattr(vos_module, voclass)):
                continue
            else:
                vo_obj._updateVOs()
                vo_data = copy.deepcopy(vo_obj._attrs)
            self._attrs[elem] = vo_data

    def getEntities(
        self,
        entities,
        config_entities=None,
        detailsLevel="BASIC",
        get_entities=False,
        debug=False,
    ):
        """Retrieve certain entities from the properties of this MambuEntity.

        Args:
          entities (list): list of strings with the name of the entity to
                           retrieve
          config_entities (list of tuples): list of tuples with information
                                            of the entity and property to
                                            instantiate
                     :py:obj:`MambuPy.api.mambustruct.MambuStruct._entities`
          detailsLevel (str): "BASIC" or "FULL" for the retrieved entities
          get_entities (bool): should MambuPy automatically instantiate
                               other MambuPy entities found inside the
                               retrieved entities?
          debug (bool): print debugging info

        Returns:
          MambuPyObject (obj): instantiation of the object from Mambu
        """
        if not config_entities:
            config_entities = self._entities
        ents = []  # do not be hasty, that is my motto
        ent_kids = []
        for entity in entities:
            entwife = [enti for enti in config_entities if enti[2] == entity]
            if len(entwife) == 0:  # you see, we lost the entwives
                if entity in self._attrs and isinstance(
                    self._attrs[entity], list
                ):  # let the ents undertake the quest to look for the entwives!
                    for entling in [
                        entkid
                        for entkid in self._attrs[entity]
                        if isinstance(entkid, MambuStruct)
                    ]:
                        ent_kids.append(
                            entling._assignEntObjs(
                                entities=entling._entities,
                                detailsLevel=detailsLevel,
                                get_entities=get_entities,
                                debug=debug,
                            )
                        )
                else:
                    raise MambuPyError(
                        "The name {} is not part of the nested entities".format(entity)
                    )
            else:  # come back to me and say my land is best!
                ents.extend(entwife)

        if len(ents) > 0:
            return self._assignEntObjs(
                entities=ents,
                detailsLevel=detailsLevel,
                get_entities=get_entities,
                debug=debug,
            )
        return ent_kids

    def __get_enc_key(self, proprty):
        """Gets a property from the obj, supposedly holding an encodedKey to
        return.

        Args:
          proprty (str): a property to get, supposedly an encodedKey.

        Returns:
          encodedKey (str) or None
        """
        try:
            enc_key = getattr(self, proprty)
            if not isinstance(enc_key, str) and not isinstance(enc_key, list):
                try:
                    enc_key = enc_key["encodedKey"]
                except KeyError:
                    enc_key = enc_key["id"]
        except AttributeError:
            try:
                prop, index, key = proprty.split("/")
                enc_key = getattr(self, prop)[int(index)][key]
            except ValueError:
                return None
        except KeyError:
            return None

        return enc_key

    def __instance_entity_obj(self, encoded_key, ent_mod, ent_class, **kwargs):
        """Instantiates a single MambuPy object calling its get method.
        If the object doesn't supports detailsLevel (MambuProduct),
           omit it.
        Args:
             encoded_key (str): encoded key of the entity to retrieve
                                from Mambu
             ent_mod (obj): module holding the class to instantiate
             ent_class (str): class to instantiate
             kwargs (dict): extra parameters for the get method
        Returns:
             MambuPyObject (obj): instantiation of the object from Mambu
        """
        try:
            try:
                return getattr(ent_mod, ent_class).get(encoded_key, **kwargs)
            except TypeError:
                kwargs.pop("detailsLevel")
                return getattr(ent_mod, ent_class).get(encoded_key, **kwargs)
        except AttributeError:
            return

    def __assign_new_property(self, new_property, ent_item, default_prop_path):
        """Assigns the value ent_item to a new property in the obj.

        Args:
          new_property (str): a new property to be assigned in  the obj.
          ent_item (obj): the value to assign
          default_prop_path (str): a / divided string with the path of the
                                   property, in case that new_property doesn't
                                   exists originally in the obj
        """
        try:
            entity = self._attrs[new_property]
            if entity.__class__.__name__ == self._cf_class.__name__:
                self._attrs[new_property].value = ent_item
                try:
                    proprty, index, key = entity.path[2:].split("/")
                    getattr(self, proprty)[int(index)][key] = ent_item
                except ValueError:
                    pass
            self["" + new_property] = entity
            self[new_property] = ent_item
        except KeyError:
            try:
                proprty, index, key = default_prop_path.split("/")
                getattr(self, proprty)[int(index)][key] = ent_item
                setattr(self, key + "_" + index, ent_item)
            except ValueError:
                self[new_property] = ent_item

    def _assignEntObjs(
        self, entities=None, detailsLevel="BASIC", get_entities=False, debug=False
    ):
        """Loops entities list of tuples to instantiate MambuPy entities from Mambu.

        End result: new properties will appear on the MambuPy object other
        :py:obj:`MambuPy.api.entities.MambuEntity` objects retrieved from Mambu

        Args:
          entities (list): list of tuples with information of the entity and
                           property to instantiate. Look at
                           :py:obj:`MambuPy.api.mambustruct.MambuStruct._entities`
          detailsLevel (str): "BASIC" or "FULL" for the retrieved entities
          get_entities (bool): should MambuPy automatically instantiate
                               other MambuPy entities found inside the
                               retrieved entities?
          debug (bool): print debugging info
        """
        if entities is None:
            entities = self._entities

        instances = []
        for encodedKey, ent_path, new_property in entities:
            ent_module = ".".join(ent_path.split(".")[:-1])
            ent_class = ent_path.split(".")[-1]
            try:
                ent_mod = import_module("." + ent_module, "mambupy.api")
            except ModuleNotFoundError:
                ent_mod = import_module(ent_module)

            enc_key = self.__get_enc_key(encodedKey)
            if not enc_key:
                continue

            if isinstance(enc_key, list):
                ent_item = []
                for item in enc_key:
                    ent_item.append(
                        self.__instance_entity_obj(
                            item,
                            ent_mod,
                            ent_class,
                            detailsLevel=detailsLevel,
                            get_entities=get_entities,
                            debug=debug,
                        )
                    )
            else:
                ent_item = self.__instance_entity_obj(
                    enc_key,
                    ent_mod,
                    ent_class,
                    detailsLevel=detailsLevel,
                    get_entities=get_entities,
                    debug=debug,
                )

            self.__assign_new_property(new_property, ent_item, encodedKey)

            instances.append(ent_item)

        return instances



"""Schema tables for Mambu Addresses.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary
"""

from sqlalchemy import Column, DateTime, ForeignKey, Integer, Numeric, String
from sqlalchemy.orm import relationship

from . import schema_orm as orm

dbname = orm.dbname
session = orm.session
Base = orm.Base


class Address(Base):
    """Adress table."""

    __tablename__ = "address"
    __table_args__ = {"schema": dbname, "keep_existing": True}

    # Columns
    encodedKey = Column(String, primary_key=True)
    parentKey = Column(String)  # redundant with same-as-RESTAPI-case
    city = Column(String)
    country = Column(String)
    line1 = Column(String)
    line2 = Column(String)
    postcode = Column(String)
    region = Column(String)

    client = relationship(
        "Client",
        back_populates="addresses",
        foreign_keys="Address.parentKey",
        primaryjoin="Address.parentKey == Client.encodedKey",
    )
    branch = relationship(
        "Branch",
        back_populates="addresses",
        foreign_keys="Address.parentKey",
        primaryjoin="Address.parentKey == Branch.encodedKey",
    )
    group = relationship(
        "Group",
        back_populates="addresses",
        foreign_keys="Address.parentKey",
        primaryjoin="Address.parentKey == Group.encodedKey",
    )

    @property
    def address(self):
        return "{}, {}, {}, {}, {}, {}".format(
            self.line1, self.line2, self.region, self.city, self.country, self.postcode
        )

    def __repr__(self):
        return "<Address(address={})>".format(self.address)



"""MambuCustomField entity: a MambuEntity struct for Custom Fields.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary
"""

from .entities import MambuEntity
from MambuPy.mambuutil import MambuPyError


class MambuCustomField(MambuEntity):
    """MambuCustomField entity"""

    _prefix = "customfields"
    """prefix constant for connections to Mambu"""


class MambuCustomFieldSet(MambuEntity):
    """MambuCustomFieldSet entity"""

    _prefix = "customfieldsets"
    """prefix constant for connections to Mambu"""

    _availableFor = [
        "CLIENT",
        "GROUP",
        "CREDIT_ARRANGEMENT",
        "LOAN_ACCOUNT",
        "GUARANTOR",
        "ASSET",
        "DEPOSIT_ACCOUNT",
        "DEPOSIT_PRODUCT",
        "TRANSACTION_CHANNEL",
        "TRANSACTION_TYPE",
        "BRANCH",
        "CENTRE",
        "USER",
    ]
    """for which entites does a Set may be available for"""

    @classmethod
    def get_all(cls, *args, **kwargs):
        """get_all, customfieldsets, filtering allowed

        Args:
          offset (int): pagination, index to start searching
          limit (int): pagination, number of elements to retrieve
          paginationDetails (str ON/OFF): ask for details on pagination
          detailsLevel (str BASIC/FULL): ask for extra details or not
          kwargs (dict): extra parameters that a customfieldset may receive in
                         its get_all method. May include a user, pwd and url to
                         connect to Mambu.
        Returns:
          list of instances of a customfieldset with data from Mambu
        """
        if "filters" in kwargs:
            raise MambuPyError("filters is not allowed on MambuCustomFieldSet")
        if "sortBy" in kwargs:
            raise MambuPyError("sortBy is not allowed on MambuCustomFieldSet")
        if (
            "availableFor" in kwargs
            and kwargs["availableFor"]
            and kwargs["availableFor"] not in cls._availableFor
        ):
            raise MambuPyError(
                "key {} not in allowed _availableFor: {}".format(
                    kwargs["availableFor"], cls._availableFor
                )
            )

        return super().get_all(*args, **kwargs)



"""Connectors to Mambu.

Currently supports REST.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary
"""

from abc import ABC, abstractmethod


class MambuConnector(ABC):
    """Interface for Connectors"""


class MambuConnectorReader(ABC):
    """Interface for Readers.

    A Reader supports the followint operations:

    - get (gets a single entity)
    - get_all (gets several entities, filtering allowed)
    - search (gets several entities, advanced filtering)
    - get_documents_metadata (gets the metadata of documents attached to some
                              entity)
    """

    @abstractmethod
    def mambu_get(self, entid, prefix, detailsLevel="BASIC"):
        """get, a single entity, identified by its entid.

        Args:
          entid (str): ID for the entity
          prefix (str): entity's URL prefix
          detailsLevel (str BASIC/FULL): ask for extra details or not
        """
        raise NotImplementedError

    @abstractmethod
    def mambu_get_all(
        self,
        prefix,
        filters=None,
        offset=None,
        limit=None,
        paginationDetails="OFF",
        detailsLevel="BASIC",
        sortBy=None,
        **kwargs
    ):
        """get_all, several entities, filtering allowed

        Args:
          prefix (str): entity's URL prefix
          filters (dict): key-value filters (depends on each entity)
          offset (int): pagination, index to start searching
          limit (int): pagination, number of elements to retrieve
          paginationDetails (str ON/OFF): ask for details on pagination
          detailsLevel (str BASIC/FULL): ask for extra details or not
          sortBy (str): ``field1:ASC,field2:DESC``, sorting criteria for
                        results
          kwargs (dict): extra parameters that a specific entity may receive in
                         its get_all method
        """
        raise NotImplementedError

    @abstractmethod
    def mambu_search(
        self,
        prefix,
        filterCriteria=None,
        sortingCriteria=None,
        offset=None,
        limit=None,
        paginationDetails="OFF",
        detailsLevel="BASIC",
    ):
        """search, several entities, filtering criteria allowed

        Args:
          filterCriteria (list of dicts): fields according to
                              LoanAccountFilterCriteria schema
          sortingCriteria (dict): fields according to
                            LoanAccountSortingCriteria
          offset (int): pagination, index to start searching
          limit (int): pagination, number of elements to retrieve
          paginationDetails (str ON/OFF): ask for details on pagination
          detailsLevel (str BASIC/FULL): ask for extra details or not
        """
        raise NotImplementedError

    @abstractmethod
    def mambu_get_documents_metadata(
        self,
        entid,
        owner_type,
        offset=None,
        limit=None,
        paginationDetails="OFF",
    ):
        """Gets metadata for all the documents attached to an entity

        Args:
          entid (str): the id or encoded key of the entity owning the document
          owner_type (str): the type of the owner of the document
          offset (int): pagination, index to start searching
          limit (int): pagination, number of elements to retrieve
          paginationDetails (str ON/OFF): ask for details on pagination
        """
        raise NotImplementedError

    @abstractmethod
    def mambu_loanaccount_getSchedule(self, loanid):
        """Retrieves the installments schedule

        Args:
          loanid (str): the id or encoded key of the loan account
        """
        raise NotImplementedError

    @abstractmethod
    def mambu_get_customfield(self, customfieldid):
        """Retrieves a Custom Field.

        Args:
          customfieldid (str): the id or encoded key of the custom field
        """
        raise NotImplementedError

    @abstractmethod
    def mambu_get_comments(self, owner_id, owner_type):
        """Retrieves the comments of entity with owner_id.

        Args:
          owner_id (str): the id or encoded key of the owner
          owner_type (str): the type of the entity owning the comments
        """
        raise NotImplementedError


class MambuConnectorWriter(ABC):
    """Interface for Writers.

    A Reader supports the followint operations:

    - update (updates an entity)
    - create (creates an entity)
    - patch (patches an entity)
    - upload_document (gets a single entity)
    - delete_document (gets several entities, filtering allowed)
    """

    @abstractmethod
    def mambu_update(self, entid, prefix, attrs):
        """updates a mambu entity

        Args:
          entid (str): the id or encoded key of the entity owning the document
          prefix (str): entity's URL prefix
          attrs (dict): entity to be updated, complying with Mambu's schemas
        """
        raise NotImplementedError

    @abstractmethod
    def mambu_create(self, prefix, attrs):
        """creates a mambu entity

        Args:
          prefix (str): entity's URL prefix
          attrs (dict): entity to be created, complying with Mambu's schemas
        """
        raise NotImplementedError

    @abstractmethod
    def mambu_patch(self, entid, prefix, fields):
        """patches certain parts of a mambu entity"""
        raise NotImplementedError

    @abstractmethod
    def mambu_delete(self, entid, prefix):
        """deletes a mambu entity"""
        raise NotImplementedError

    @abstractmethod
    def mambu_upload_document(self, owner_type, entid, filename, name, notes):
        """uploads an attachment to this entity

        Args:
          owner_type (str): the type of the owner of the document
          entid (str): the id or encoded key of the entity owning the document
          filename (str): path and filename of file to upload as attachment
          name (str): name to assign to the attached file in Mambu
          notes (str): notes to associate to the attached file in Mambu
        """
        raise NotImplementedError

    @abstractmethod
    def mambu_delete_document(self, documentId):
        """deletes an attachment by its documentId

        Args:
          documentId (str): the id or encodedkey of the document to be deleted
        """
        raise NotImplementedError

    @abstractmethod
    def mambu_change_state(self, entid, prefix, action, notes):
        """change state of mambu entity

        Args:
          entid (str): the id or encoded key of the entity owning the document
          prefix (str): entity's URL prefix
          action (str): specify the action state
          notes (str): notes to associate to the change of status
        """
        raise NotImplementedError

    @abstractmethod
    def mambu_comment(self, owner_id, owner_type, text):
        """Comments an entity with owner_id.

        Args:
          owner_id (str): the id or encoded key of the owner
          owner_type (str): the type of the entity owning the comments
          text (str): the text of the comment
        """
        raise NotImplementedError

    @abstractmethod
    def mambu_make_disbursement(
        self, loan_id, firstRepaymentDate, notes, valueDate, allowed_fields, **kwargs
    ):
        """Make a disbursement transacton on a loan account.

        Args:
          loan_id (str): loan account id to disburse
          firstRepaymentDate (str): first repayment date in ISO format
          notes (str): notes for the disbursement transaction
          valueDate (str): entry date for the disbursement transaction
          allowed_fields (list): extra fields allowed for the transaction
          kwargs (dict): key-values of extra fields for the transaction
        """
        raise NotImplementedError

    @abstractmethod
    def mambu_make_repayment(
        self, loan_id, amount, notes, valueDate, allowed_fields, **kwargs
    ):
        """Make a repayment transaction on a loan account.

        Args:
          loan_id (str): loan account id to make a repayment
          amount (float): the amount of the repayment
          notes (str): notes for the repayment transaction
          valueDate (str): entry date for the repayment transaction
          allowed_fields (list): extra fields allowed for the transaction
          kwargs (dict): key-values of extra fields for the transaction
        """
        raise NotImplementedError

    @abstractmethod
    def mambu_make_fee(
        self, loan_id, amount, installmentNumber, notes, valueDate, allowed_fields, **kwargs
    ):
        """Make a fee transaction on a loan account.

        Args:
          loan_id (str): loan account id to apply a fee
          amount (float): the amount of the fee
          installmentNumber (int): the installment number to apply the fee
          notes (str): notes for the fee transaction
          valueDate (str): entry date for the fee transaction
          allowed_fields (list): extra fields allowed for the transaction
          kwargs (dict): key-values of extra fields for the transaction
        """
        raise NotImplementedError

    @abstractmethod
    def mambu_loanaccount_writeoff(self, loanid, notes):
        """Writesoff a loan account

        Args:
          loanid (str): the id or encoded key of the loan account
          notes (str): notes to associate to the writeoff operation in Mambu
        """
        raise NotImplementedError

    @abstractmethod
    def mambu_loantransaction_adjust(self, transactionid, notes):
        """Adjust a loan transaction

        Args:
          transactionid (str): the id or encoded key of the loan transaction
          notes (str): notes to associate to the transaction adjustment in Mambu
        """
        raise NotImplementedError



"""Schema tables for Mambu Activities.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary
"""

from datetime import datetime

from sqlalchemy import (Column, DateTime, ForeignKey, Integer, Numeric, String,
                        Table, Text)
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import backref, relationship

from . import schema_orm as orm
from .schema_branches import Branch
from .schema_clients import Client
from .schema_groups import Group
from .schema_loans import LoanAccount
from .schema_users import User

dbname = orm.dbname
session = orm.session
Base = orm.Base


class Task(Base):
    """Task table."""

    __tablename__ = "task"
    __table_args__ = {"schema": dbname, "keep_existing": True}

    # Columns
    encodedKey = Column(String, primary_key=True)
    creationDate = Column(DateTime)
    completionDate = Column(DateTime)
    description = Column(Text)
    dueDate = Column(DateTime)
    id = Column(Integer)
    lastModifiedDate = Column(DateTime)
    status = Column(String)
    title = Column(String)
    taskLinkType = Column(String)

    # Relationships
    assignedUserKey = Column(String, ForeignKey(User.encodedKey))
    createdByUserKey = Column(String, ForeignKey(User.encodedKey))

    createdByUser = relationship(
        User,
        primaryjoin="User.encodedKey==Task.createdByUserKey",
        back_populates="createdTasks",
    )
    assignedUser = relationship(
        User,
        primaryjoin="User.encodedKey==Task.assignedUserKey",
        back_populates="assignedTasks",
    )

    taskLinkKey = Column(String)
    link_group = relationship(
        "Group",
        back_populates="tasks",
        foreign_keys="Task.taskLinkKey",
        primaryjoin="Task.taskLinkKey == Group.encodedKey",
    )
    link_loan = relationship(
        "LoanAccount",
        back_populates="tasks",
        foreign_keys="Task.taskLinkKey",
        primaryjoin="Task.taskLinkKey == LoanAccount.encodedKey",
    )

    @hybrid_property
    def daysUntilDue(self):
        """Days until due date"""
        diffDays = self.dueDate - datetime.now()
        return diffDays

    def __repr__(self):
        return "<Task(taskLinkType=%s, dueDate=%s)>" % (
            self.taskLinkType,
            self.dueDate.strftime("%Y%m%d"),
        )



"""Deactivate multiple users in a csv list."""
import argparse
import csv

from mambupy.api import mambugroup
from mambupy.api import mambuloan
from mambupy.api import mambuuser


def verify_loans(username):
    """Verify the loans wallet of given username.

    If user does meet the conditions:
    User has loans with state equal to 'ACTIVE' or 'ACTIVE_IN_ARREARS' and its
    loans balance are mayor than 0 the method returns False, if not returns True.

    Args:
        username (str): the username of the user to deactivate.

    Returns:
        (bool): depending on user's loans satisfying or not the given conditions.
    """

    loans = mambuloan.MambuLoan().get_all(
        filters={'creditOfficerUsername': username},
        detailsLevel='FULL')

    balanceSum = 0
    accountState = False
    for loan in loans:
        balanceSum += loan.balances['principalBalance'] \
            + loan.balances['interestBalance'] \
            + loan.balances['feesBalance'] \
            + loan.balances['penaltyBalance']
        if loan.accountState in ['ACTIVE', 'ACTIVE_IN_ARREARS']:
            accountState = True

    if balanceSum > 0 or accountState:
        return False
    else:
        return True


def deactivate_user(username):
    """Deactivate the given username.

    Calls to verify_loans to check if the user is deactivatable. If it not
    returns False.

    Unassign user's groups by changing the assignedCentreKey and assignedUserKey
    fields to None for each group that user has.

    Change userState field value to INACTIVE to deactivate user and returns True.

    Args:
        username (str): Username of the user to deactivate.

    Returns:
        (bool): depending on verify_loan's verification.
    """
    if verify_loans(username=username):

        groups = mambugroup.MambuGroup().get_all(
            filters={'creditOfficerUsername': username},
            detailsLevel='FULL')

        for group in groups:
            del group['assignedUserKey']
            group.patch(autodetect_remove=True)

        user = mambuuser.MambuUser().get(
            username, detailsLevel="FULL")
        user.userState = 'INACTIVE'
        user.patch(['userState'])

        return True

    else:
        return False


def main(csv_file):
    users = []
    with open(csv_file) as f:
        reader = csv.reader(f, delimiter=',')
        for us in reader:
            users.append(tuple(us))

    usersCanNotDeactivate = []
    for user in users:
        if deactivate_user(user):
            print('{0} user deactivated.'.format(user))
        else:
            usersCanNotDeactivate.append(user)

    if usersCanNotDeactivate:
        print("Users that could not be deactivated:")
        print(usersCanNotDeactivate)


if __name__ == "__main__":  # pragma: no cover
    argparser = argparse.ArgumentParser(
        description='Deactivates users from a list.'
    )
    argparser.add_argument(
        'users_list',
        help='CSV holding the usernames to deactivate.'
        '(col1: username)'
    )

    args = argparser.parse_args()

    csv_file = args.users_list

    main(csv_file)



"""MambuProduct entity: a MambuEntity struct for Products.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary
"""

from .entities import (
    MambuEntity,
    MambuEntityWritable,
    MambuEntityAttachable,
    MambuEntityCommentable,
)


class MambuProduct(
    MambuEntity, MambuEntityWritable, MambuEntityAttachable, MambuEntityCommentable
):
    """MambuProduct entity"""

    _prefix = "loanproducts"
    """prefix constant for connections to Mambu"""

    _filter_keys = []
    """allowed filters for get_all filtering"""

    _sortBy_fields = [
        "creationDate",
        "lastModifiedDate",
        "id",
        "productName",
    ]
    """allowed fields for get_all sorting"""

    _ownerType = "LOAN_PRODUCT"
    """owner type of this entity"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._attachments = {}

    @classmethod
    def get(cls, entid, get_entities=False, **kwargs):
        """get, a single entity, identified by its entid.

        Args:
          entid (str): ID for the entity
          get_entities (bool): should MambuPy automatically instantiate other
                               MambuPy entities found inside the retrieved
                               entity?

        Returns:
          instance of an entity with data from Mambu
        """
        if "detailsLevel" in kwargs:
            kwargs.pop("detailsLevel")
        return super().get(
            entid, detailsLevel="FULL", get_entities=get_entities, **kwargs
        )

    def refresh(self):
        """get again this single entity, identified by its entid.

        Updates _attrs with responded data. Loses any change on _attrs that
        overlaps with anything from Mambu. Leaves alone any other properties
        that don't come in the response.
        """
        super().refresh(detailsLevel="FULL")

    @classmethod
    def get_all(
        cls, filters=None, offset=None, limit=None, paginationDetails="OFF", sortBy=None, **kwargs
    ):
        """get_all, several entities, filtering allowed

        Args:
          filters (dict): key-value filters, dependes on each entity
                          (keys must be one of the _filter_keys)
          offset (int): pagination, index to start searching
          limit (int): pagination, number of elements to retrieve
          paginationDetails (str ON/OFF): ask for details on pagination
          sortBy (str): ``field1:ASC,field2:DESC``, sorting criteria for
                        results (fields must be one of the _sortBy_fields)

        Returns:
          list of instances of an entity with data from Mambu
        """
        if "detailsLevel" in kwargs:
            kwargs.pop("detailsLevel")
        return super().get_all(
            filters, offset, limit, paginationDetails, detailsLevel="FULL", sortBy=sortBy, **kwargs
        )



"""MambuGroup entity: a MambuEntity struct for Groups of clients.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary
"""

from .entities import (
    MambuEntity,
    MambuEntityWritable,
    MambuEntityAttachable,
    MambuEntitySearchable,
    MambuEntityCommentable,
)
from .interfaces import MambuHolder


class MambuGroup(
    MambuEntity,
    MambuEntityWritable,
    MambuEntityAttachable,
    MambuEntitySearchable,
    MambuEntityCommentable,
    MambuHolder,
):
    """MambuGroup entity"""

    _prefix = "groups"
    """prefix constant for connections to Mambu"""

    _default_tzattrs = {
        "addresses": [{}],
        "groupMembers": [],
    }

    _filter_keys = ["branchId", "centreId", "creditOfficerUsername"]
    """allowed filters for get_all filtering"""

    _sortBy_fields = [
        "creationDate",
        "lastModifiedDate",
        "groupName",
    ]
    """allowed fields for get_all sorting"""

    _ownerType = "GROUP"
    """owner type of this entity"""

    _vos = [("addresses", "MambuAddress"), ("groupMembers", "MambuGroupMember")]
    """2-tuples of elements and Value Objects"""

    _entities = [
        ("assignedUserKey", "mambuuser.MambuUser", "assignedUser"),
        ("assignedBranchKey", "mambubranch.MambuBranch", "assignedBranch"),
        ("assignedCentreKey", "mambucentre.MambuCentre", "assignedCentre"),
    ]
    """3-tuples of elements and Mambu Entities"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._attachments = {}



"""This file imports all ORM schemas.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary

..
  ejemplos de intersphinx...
  :py:class:`zipfile.ZipFile`
  :py:func:`sqlalchemy.create_engine`
  :py:func:`sqlalchemy.create_engine`

To use the ORM you MUST import schema_mambu, if you try to use the schemas
individualy you will get some relationship() errors, for example,
if you use schema_loans, you will get an error like this::

 from mambupy.schema_loans import *
 loan = session.query(LoanAccount).filter(LoanAccount.id=="12345").one()

... When initializing mapper Mapper|Branch|branch, expression ...

Instead, if you use schema_mambu the query executes correctly::

 from mambupy.schema_mambu import *
 loan = session.query(LoanAccount).filter(LoanAccount.id=="12345").one()
"""

from .schema_activities import *
from .schema_addresses import *
from .schema_branches import *
from .schema_centres import *
from .schema_clients import *
from .schema_customfields import *
from .schema_groups import *
from .schema_loans import *
from .schema_tasks import *
from .schema_users import *



"""Mambu configuration file.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary

IMPORTANT: be careful of who can read/import this file. It may contain
sensitive data to connect to your Mambu instance, or to your DB server.

Supports ConfigParser to read a config file (in INI format) from
$HOME/.mambupy.rc, or /etc/mambupy.rc if the previous one not found. If you
send a --mambupy_rcfile it overrides this previous behavior using the RC file
wherever you say it is.

You may edit the default_configs dictionary to have some defaults defined
even if no RC file is found (NOT RECOMMENDED)

RC files must have the following format:

.. code-block:: ini

    [API]
    apiurl=url_to_mambu_domain
    apiuser=API_user
    apipwd=API_password
    apipagination=API_pagination_limit
    activate_request_session_objects=API_persistent_session
    [DB]
    dbname=Database_name
    dbuser=Database_user
    dbpwd=Database_password
    dbhost=Database_host
    dbport=Database_port
    dbeng=Database_engine
    [MAMBUPY]
    loggingconf=Logging_conf_path

You can use environment variables to override the previous behaviour for any
option. The variable must be named as follows::

    MAMBUPY_UPPERCASEOPTION

You can use command line arguments to override the previous behaviour for any
option. The command line argument must be named as follows::

    mambupy_lowercaseoption

IMPORTANT!!! if you use argparse anywhere on your own programs using MambuPy,
then you must parse using the parse_known_args() method from your parser.
Otherwise, your programs won't work by not recognizing MambuPy's own command
line arguments.

So, the configuration hierarchy goes like this::

    0) Command line arguments (overrides any config on the following)
      1) Environment variables (overrides any config on the following)
        2) RC file sent via --mambupy_rcfile command line argument
        3) RC file at $HOME/.mambupy.rc (if previous one not found)
        4) RC file at /etc/mambupy.rc (if previous one not found)
          5) values of default_configs dictionary, if any config is absent
             on any of the previous, this is used to default it

Note for Mambu SysAdmins: please correctly configure your API user, with
right permissions that allow your users to access the data you wish to
make available but not to make any changes/access any information you
don't wish to be accessed. Please mind that this may restrict some of
the Mambu objects at the MambuPy package to don't be fully available to
python users of this library.

Note for DBAs: to increase security, configure your DB user to access
the Mambu DB backup so that it cannot access any other DB tables you
don't wish to be accessed by any other user

The api* configs are Mambu API configurations, you must have a valid API
user on your Mambu instance

The db* configs refer to Database configurations which hold a backup of
your Mambu Database on some server.

The activate_request_session_objects config enables or disables HTTP persistent sessions
for API requests. When enabled, it will reuse the same TCP connection for
multiple requests, improving performance. Default is False.
"""

default_configs = {
    # Mambu API configurations
    "apiurl": "domain.mambu.com",
    "apiuser": "mambu_api_user",
    "apipwd": "mambu_api_password",
    "apipagination": "50",
    "activate_request_session_objects": "False",
    # Mambu DB configurations
    "dbname": "mambu_db",
    "dbuser": "mambu_db_user",
    "dbpwd": "mambu_db_pwd",
    "dbhost": "localhost",
    "dbport": "3306",
    "dbeng": "mysql",
    # MambuPy configurations
    "loggingconf": "",  # empty string means default logging, path to a yaml for logging configuration
}
""" Defaults dictionary for the options configured here.

You may edit this to your own liking. But beware of pricking eyes!
"""

import os
import configparser
from configparser import NoSectionError

config = configparser.ConfigParser(defaults=default_configs)

# argparse for command line arguments overriding
import argparse

argparser = argparse.ArgumentParser()
argparser.add_argument("--mambupy_rcfile")
argparser.add_argument("--mambupy_apiurl")
argparser.add_argument("--mambupy_apiuser")
argparser.add_argument("--mambupy_apipwd")
argparser.add_argument("--mambupy_apipagination")
argparser.add_argument("--mambupy_activate_request_session_objects")
argparser.add_argument("--mambupy_dbname")
argparser.add_argument("--mambupy_dbuser")
argparser.add_argument("--mambupy_dbpwd")
argparser.add_argument("--mambupy_dbhost")
argparser.add_argument("--mambupy_dbport")
argparser.add_argument("--mambupy_dbeng")
argparser.add_argument("--mambupy_loggingconf")
args, unknown = argparser.parse_known_args()


def get_conf(conf, sect, opt):
    """Gets a config 'opt' from 'conf' file, under section 'sect'.

    If no 'opt' exists under 'sect', it looks for option on the default_configs
    dictionary

    If there exists an environmental variable named MAMBUPY_{upper_case_opt},
    it overrides whatever the conf files or default_configs dict says.

    But if you send a command line argument named mambupy_{lower_case_opt},
    it overrides anything else.

    Args:
     conf (ConfigParser): ConfigParser that reads from certain config file (INI
                          format)
     sect (string): section under the config file
     opt (string): option to read

    Returns:
     string: configuration option. If not found on conf, returns a value from
             default_configs dict. If environmental variable exists with name
             MAMBUPY_{upper_case_opt} it overrides anything else
    """
    argu = getattr(args, "mambupy_" + opt.lower())

    if not argu:
        envir = os.environ.get("MAMBUPY_" + opt.upper())
        if not envir:
            try:
                return conf.get(sect, opt)
            except NoSectionError:
                return default_configs[opt]
        return envir
    return argu


# Read config from --mambupy_rcfile, or $HOME/.mambupy.rc if not found, or
# /etc/mambupy.rc if not found
if not args.mambupy_rcfile or not config.read(args.mambupy_rcfile):
    if "HOME" not in os.environ or not config.read(os.environ["HOME"] + "/.mambupy.rc"):
        config.read("/etc/mambupy.rc")

# Read configs from config file
apiurl = get_conf(config, "API", "apiurl")
"""URL to access Mambu API"""
apiuser = get_conf(config, "API", "apiuser")
"""Username to access Mambu API"""
apipwd = get_conf(config, "API", "apipwd")
"""Password to access Mambu API"""
apipagination = get_conf(config, "API", "apipagination")
"""Pagination default limit for requests to Mambu API"""
activate_request_session_objects = get_conf(config, "API", "activate_request_session_objects")
"""Whether to use persistent HTTP sessions for API requests"""
dbname = get_conf(config, "DB", "dbname")
"""Name of the DB with a backup of Mambu's DB"""
dbuser = get_conf(config, "DB", "dbuser")
"""Username to connect to the Mambu's DB backup"""
dbpwd = get_conf(config, "DB", "dbpwd")
"""Password to connect to the Mambu's DB backup"""
dbhost = get_conf(config, "DB", "dbhost")
"""Host where the Mambu's DB backup lives"""
dbport = get_conf(config, "DB", "dbport")
"""Port to connect to the host of the Mambu's DB backup"""
dbeng = get_conf(config, "DB", "dbeng")
"""DB engine for the Mambu's DB backup"""
loggingconf = get_conf(config, "MAMBUPY", "loggingconf")
"""Directory where to store logs"""



"""Make Dummy ORM objects.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary

Sample dummies are created here. Please DO NOT TRY to persist them to the DB.
"""

from datetime import datetime as dt


def make_dummy(
    instance,
    relations={},
    datetime_default=dt.strptime("1901-01-01", "%Y-%m-%d"),
    varchar_default="",
    integer_default=0,
    numeric_default=0.0,
    *args,
    **kwargs
):
    """Make an instance to look like an empty dummy.

    Every field of the table is set with zeroes/empty strings.
    Date fields are set to 01/01/1901.

    * relations is a dictionary to set properties for relationships
      on the instance.

    The keys of the relations dictionary are the name of the fields to
    be set at the instance.

    The values of the relations dictionary must be 2 dimension tuples:

    - first element will be the element to be related.

    - second element will be the name of the backref set on the previous
      first element of the tuple, as a list containing the instance.

    If you wish that no backref to be set you may use any invalid value
    for a property name, anything other than a string, for example a
    number. Preferably, use None.

    * datetime_default is the datetime object to use for init datetime
      fields. Defaults to 01-01-1901

    * varchar_default is the string object to use for init varchar
      fields. Defaults to ""

    * integer_default is the int object to use for init integer
      fields. Defaults to 0

    * numeric_default is the float object to use for init numeric
      fields. Defaults to 0.0

    * kwargs may have the name of a certain field you wish to initialize
      with a value other than the given by the init_data dicionary. If you
      give an unexistent name on the columns of the table, they are safely
      ignored.

    .. todo:: further field types may be set at the init_data dictionary."""

    # init_data knows how to put an init value depending on data type
    init_data = {
        "DATETIME": datetime_default,
        "VARCHAR": varchar_default,
        "INTEGER": integer_default,
        "NUMERIC(50, 10)": numeric_default,
        "TEXT": varchar_default,
    }

    # the type of the instance is the SQLAlchemy Table
    table = type(instance)

    for col in table.__table__.columns:
        # declarative base tables have a columns property useful for reflection
        try:
            setattr(instance, col.name, kwargs[col.name])
        except KeyError:
            setattr(instance, col.name, init_data[str(col.type)])

    for k, v in relations.items():
        # set the relationship property with the first element of the tuple
        setattr(instance, k, v[0])
        # try:
        #     # set the relationship backref on the first element of the tuple
        #     # with a property named according to the second element of the
        #     # tuple, pointing to a list with the instance itself (assumes a
        #     # one-to-many relationship)
        #     # in case you don't want a backref, just send a None as v[1]
        #     try:
        #         getattr(v[0], v[1]).append(instance)
        #     except:
        #         setattr(v[0], v[1], [ instance ])
        # except:
        #     pass

    return instance


# Sample DUMMIES

# Some sample dummies with all the current tables mapped on MambuPy.
# When creating new tables, please add them here.

# Current relationships are also dummified here, please add new relations when
# existing tables are updated with further relationships
from .schema_mambu import *

# branches
branch = make_dummy(Branch())

# centres
centre = make_dummy(Centre())

# users
role = make_dummy(Role())
user = make_dummy(
    User(),
    relations={
        "branch": (branch, "users"),
        "role": (role, "users"),
    },
)

# groups
group = make_dummy(
    Group(),
    relations={
        "branch": (branch, "groups"),
        "centre": (centre, "groups"),
        "user": (user, "groups"),
    },
)

# clients
iddoc = make_dummy(IdentificationDocument())
client = make_dummy(
    Client(),
    relations={
        "branch": (branch, "clients"),
        "groups": ([group], "clients"),
        "identificationDocuments": ([iddoc], "client"),
    },
)

# loan accounts
loanproduct = make_dummy(LoanProduct())
disbursementdetails = make_dummy(DisbursementDetails())
repayment = make_dummy(Repayment())
transactionchannel = make_dummy(TransactionChannel())
transactiondetails = make_dummy(
    TransactionDetails(), relations={"transactionChannel": (transactionchannel, None)}
)
loantransaction = make_dummy(
    LoanTransaction(), relations={"transactionDetails": (transactiondetails, None)}
)
loanaccount = make_dummy(
    LoanAccount(),
    relations={
        "product": (loanproduct, None),
        "disbursementDetails": (disbursementdetails, None),
        "branch": (branch, None),
        "user": (user, "loans"),
        "holder_group": (group, "loans"),
        "holder_client": (client, "loans"),
        "repayments": ([repayment], "account"),
        "transactions": ([loantransaction], "account"),
    },
)

# custom fields
custominformation = make_dummy(CustomField())
customfieldvalue = make_dummy(
    CustomFieldValue(),
    relations={
        "customField": (custominformation, "customFieldValues"),
        "branch": (branch, "customInformation"),
        "user": (user, "customInformation"),
        "group": (group, "customInformation"),
        "client": (client, "customInformation"),
        "loan": (loanaccount, "customInformation"),
    },
)

# addresses
address = make_dummy(
    Address(),
    relations={
        "branch": (branch, "addresses"),
        "group": (group, "addresses"),
        "client": (client, "addresses"),
    },
)

# activities
activity = make_dummy(
    Activity(),
    relations={
        "loan": (loanaccount, "activities"),
        "branch": (branch, "activities"),
        "client": (client, "activities"),
        "group": (group, "activities"),
        "user": (user, "activities"),
        "assignedUser": (user, "assignedActivities"),
    },
)
# tasks
tasks = make_dummy(
    Task(),
    relations={
        "createdByUser": (user, "createdTasks"),
        "assignedUser": (user, "assignedTasks"),
        "link_group": (group, "tasks"),
        "link_loan": (loanaccount, "tasks"),
    },
)
